@@ -13,6 +13,8 @@ where
 
 #include "HsVersions.h"
 
+import GhcPrelude
+
 import Vectorise.Type.Type
 import Vectorise.Var
 import Vectorise.Convert
@@ -44,10 +46,6 @@ import Outputable
 import FastString
 import DynFlags
 import Util
-import UniqDFM (udfmToUfm)
-#if __GLASGOW_HASKELL__ < 709
-import MonadUtils
-#endif
 
 import Control.Monad
 import Data.Maybe
@@ -294,7 +292,7 @@ liftSimpleAndCase aexpr = liftSimple aexpr
 
 liftSimple :: CoreExprWithVectInfo -> VM CoreExprWithVectInfo
 liftSimple ((fvs, vi), AnnVar v)
-  | v `elemVarSet` fvs                -- special case to avoid producing: (\v -> v) v
+  | v `elemDVarSet` fvs               -- special case to avoid producing: (\v -> v) v
   && not (isToplevel v)               --   NB: if 'v' not free or is toplevel, we must get the 'VIEncaps'
   = return $ ((fvs, vi), AnnVar v)
 liftSimple aexpr@((fvs_orig, VISimple), expr)
@@ -306,13 +304,13 @@ liftSimple aexpr@((fvs_orig, VISimple), expr)
     ; return $ liftedExpr
     }
   where
-    vars = varSetElems fvs
-    fvs  = filterVarSet (not . isToplevel) fvs_orig -- only include 'Id's that are not toplevel
+    vars = dVarSetElems fvs
+    fvs  = filterDVarSet (not . isToplevel) fvs_orig -- only include 'Id's that are not toplevel
 
-    mkAnnLams :: [Var] -> VarSet -> AnnExpr' Var (VarSet, VectAvoidInfo) -> CoreExprWithVectInfo
-    mkAnnLams []     fvs expr = ASSERT(isEmptyVarSet fvs)
-                                ((emptyVarSet, VIEncaps), expr)
-    mkAnnLams (v:vs) fvs expr = mkAnnLams vs (fvs `delVarSet` v) (AnnLam v ((fvs, VIEncaps), expr))
+    mkAnnLams :: [Var] -> DVarSet -> AnnExpr' Var (DVarSet, VectAvoidInfo) -> CoreExprWithVectInfo
+    mkAnnLams []     fvs expr = ASSERT(isEmptyDVarSet fvs)
+                                ((emptyDVarSet, VIEncaps), expr)
+    mkAnnLams (v:vs) fvs expr = mkAnnLams vs (fvs `delDVarSet` v) (AnnLam v ((fvs, VIEncaps), expr))
 
     mkAnnApps :: CoreExprWithVectInfo -> [Var] -> CoreExprWithVectInfo
     mkAnnApps aexpr []     = aexpr
@@ -320,13 +318,14 @@ liftSimple aexpr@((fvs_orig, VISimple), expr)
 
     mkAnnApp :: CoreExprWithVectInfo -> Var -> CoreExprWithVectInfo
     mkAnnApp aexpr@((fvs, _vi), _expr) v
-      = ((fvs `extendVarSet` v, VISimple), AnnApp aexpr ((unitVarSet v, VISimple), AnnVar v))
+      = ((fvs `extendDVarSet` v, VISimple), AnnApp aexpr ((unitDVarSet v, VISimple), AnnVar v))
 liftSimple aexpr
   = pprPanic "Vectorise.Exp.liftSimple: not simple" $ ppr (deAnnotate aexpr)
 
 isToplevel :: Var -> Bool
 isToplevel v | isId v    = case realIdUnfolding v of
                              NoUnfolding                     -> False
+                             BootUnfolding                   -> False
                              OtherCon      {}                -> True
                              DFunUnfolding {}                -> True
                              CoreUnfolding {uf_is_top = top} -> top
@@ -363,7 +362,8 @@ vectExpr (_, AnnApp (_, AnnApp (_, AnnVar v) (_, AnnType ty)) err)
   | v == pAT_ERROR_ID
   = do
     { (vty, lty) <- vectAndLiftType ty
-    ; return (mkCoreApps (Var v) [Type (getRuntimeRep "vectExpr" vty), Type vty, err'], mkCoreApps (Var v) [Type lty, err'])
+    ; return (mkCoreApps (Var v) [Type (getRuntimeRep vty), Type vty, err'],
+              mkCoreApps (Var v) [Type lty, err'])
     }
   where
     err' = deAnnotate err
@@ -658,7 +658,7 @@ mkScalarFun arg_tys res_ty expr
 -- In other words, all methods in that dictionary are scalar functions — to be vectorised with
 -- 'vectScalarFun'.  The dictionary "function" itself may be a constant, though.
 --
--- NB: You may think that we could implement this function guided by the struture of the Core
+-- NB: You may think that we could implement this function guided by the structure of the Core
 --     expression of the right-hand side of the dictionary function.  We cannot proceed like this as
 --     'vectScalarDFun' must also work for *imported* dfuns, where we don't necessarily have access
 --     to the Core code of the unvectorised dfun.
@@ -766,7 +766,7 @@ vectLam inline loop_breaker expr@((fvs, _vi), AnnLam _ _)
           -- collect and vectorise all /local/ free variables
       ; vfvs <- readLEnv $ \env ->
                   [ (var, fromJust mb_vv)
-                  | var <- varSetElems fvs
+                  | var <- dVarSetElems fvs
                   , let mb_vv = lookupVarEnv (local_vars env) var
                   , isJust mb_vv         -- its local == is in local var env
                   ]
@@ -932,7 +932,7 @@ vectAlgCase tycon _ty_args scrut bndr ty alts
           vect_dc <- maybeV dataConErr (lookupDataCon dc)
           let ntag = dataConTagZ vect_dc
               tag  = mkDataConTag dflags vect_dc
-              fvs  = fvs_body `delVarSetList` bndrs
+              fvs  = fvs_body `delDVarSetList` bndrs
 
           sel_tags  <- liftM (`App` sel) (builtin (selTags arity))
           lc        <- builtin liftingContext
@@ -944,7 +944,7 @@ vectAlgCase tycon _ty_args scrut bndr ty alts
              $ do
                { binds    <- mapM (pack_var (Var lc) sel_tags tag)
                            . filter isLocalId
-                           $ varSetElems fvs
+                           $ dVarSetElems fvs
                ; traceVt "case alternative:" (ppr . deAnnotate $ body)
                ; (ve, le) <- vectExpr body
                ; return (ve, Case (elems `App` sel) lc lty
@@ -995,7 +995,7 @@ data VectAvoidInfo = VIParr       -- tree contains parallel computations
 
 -- Core expression annotated with free variables and vectorisation-specific information.
 --
-type CoreExprWithVectInfo = AnnExpr Id (VarSet, VectAvoidInfo)
+type CoreExprWithVectInfo = AnnExpr Id (DVarSet, VectAvoidInfo)
 
 -- Yield the type of an annotated core expression.
 --
@@ -1043,15 +1043,15 @@ vectAvoidInfo :: VarSet -> CoreExprWithFVs -> VM CoreExprWithVectInfo
 vectAvoidInfo pvs ce@(_, AnnVar v)
   = do
     { gpvs <- globalParallelVars
-    ; vi <- if v `elemVarSet` pvs || v `elemVarSet` gpvs
+    ; vi <- if v `elemVarSet` pvs || v `elemDVarSet` gpvs
             then return VIParr
             else vectAvoidInfoTypeOf ce
     ; viTrace ce vi []
     ; when (vi == VIParr) $
         traceVt "  reason:" $ if v `elemVarSet` pvs  then text "local"  else
-                              if v `elemVarSet` gpvs then text "global" else text "parallel type"
+                              if v `elemDVarSet` gpvs then text "global" else text "parallel type"
 
-    ; return ((udfmToUfm fvs, vi), AnnVar v)
+    ; return ((fvs, vi), AnnVar v)
     }
   where
     fvs = freeVarsOf ce
@@ -1060,7 +1060,7 @@ vectAvoidInfo _pvs ce@(_, AnnLit lit)
   = do
     { vi <- vectAvoidInfoTypeOf ce
     ; viTrace ce vi []
-    ; return ((udfmToUfm fvs, vi), AnnLit lit)
+    ; return ((fvs, vi), AnnLit lit)
     }
   where
     fvs = freeVarsOf ce
@@ -1072,7 +1072,7 @@ vectAvoidInfo pvs ce@(_, AnnApp e1 e2)
     ; eVI2 <- vectAvoidInfo pvs e2
     ; let vi = ceVI `unlessVIParrExpr` eVI1 `unlessVIParrExpr` eVI2
     -- ; viTrace ce vi [eVI1, eVI2]
-    ; return ((udfmToUfm fvs, vi), AnnApp eVI1 eVI2)
+    ; return ((fvs, vi), AnnApp eVI1 eVI2)
     }
   where
     fvs = freeVarsOf ce
@@ -1083,7 +1083,7 @@ vectAvoidInfo pvs ce@(_, AnnLam var body)
     ; varVI  <- vectAvoidInfoType $ varType var
     ; let vi = vectAvoidInfoOf bodyVI `unlessVIParr` varVI
     -- ; viTrace ce vi [bodyVI]
-    ; return ((udfmToUfm fvs, vi), AnnLam var bodyVI)
+    ; return ((fvs, vi), AnnLam var bodyVI)
     }
   where
     fvs = freeVarsOf ce
@@ -1103,7 +1103,7 @@ vectAvoidInfo pvs ce@(_, AnnLet (AnnNonRec var e) body)
         ; return (bodyVI, ceVI `unlessVIParrExpr` bodyVI)
         }
     -- ; viTrace ce vi [eVI, bodyVI]
-    ; return ((udfmToUfm fvs, vi), AnnLet (AnnNonRec var eVI) bodyVI)
+    ; return ((fvs, vi), AnnLet (AnnNonRec var eVI) bodyVI)
     }
   where
     fvs = freeVarsOf ce
@@ -1120,13 +1120,13 @@ vectAvoidInfo pvs ce@(_, AnnLet (AnnRec bnds) body)
         ; bndsVI <- mapM (vectAvoidInfoBnd extendedPvs) bnds
         ; bodyVI <- vectAvoidInfo extendedPvs body
         -- ; viTrace ce VIParr (map snd bndsVI ++ [bodyVI])
-        ; return ((udfmToUfm fvs, VIParr), AnnLet (AnnRec bndsVI) bodyVI)
+        ; return ((fvs, VIParr), AnnLet (AnnRec bndsVI) bodyVI)
         }
       else do         -- demanded bindings cannot trigger parallelism
         { bodyVI <- vectAvoidInfo pvs body
         ; let vi = ceVI `unlessVIParrExpr` bodyVI
         -- ; viTrace ce vi (map snd bndsVI ++ [bodyVI])
-        ; return ((udfmToUfm fvs, vi), AnnLet (AnnRec bndsVI) bodyVI)
+        ; return ((fvs, vi), AnnLet (AnnRec bndsVI) bodyVI)
         }
     }
   where
@@ -1147,7 +1147,7 @@ vectAvoidInfo pvs ce@(_, AnnCase e var ty alts)
     ; let alteVIs = [eVI | (_, _, eVI) <- altsVI]
           vi      =  foldl unlessVIParrExpr ceVI (eVI:alteVIs)  -- NB: same effect as in the paper
     -- ; viTrace ce vi (eVI : alteVIs)
-    ; return ((udfmToUfm fvs, vi), AnnCase eVI var ty altsVI)
+    ; return ((fvs, vi), AnnCase eVI var ty altsVI)
     }
   where
     fvs = freeVarsOf ce
@@ -1162,7 +1162,7 @@ vectAvoidInfo pvs ce@(_, AnnCase e var ty alts)
 vectAvoidInfo pvs ce@(_, AnnCast e (fvs_ann, ann))
   = do
     { eVI <- vectAvoidInfo pvs e
-    ; return ((udfmToUfm fvs, vectAvoidInfoOf eVI), AnnCast eVI ((udfmToUfm $ freeVarsOfAnn fvs_ann, VISimple), ann))
+    ; return ((fvs, vectAvoidInfoOf eVI), AnnCast eVI ((freeVarsOfAnn fvs_ann, VISimple), ann))
     }
   where
     fvs = freeVarsOf ce
@@ -1170,18 +1170,18 @@ vectAvoidInfo pvs ce@(_, AnnCast e (fvs_ann, ann))
 vectAvoidInfo pvs ce@(_, AnnTick tick e)
   = do
     { eVI <- vectAvoidInfo pvs e
-    ; return ((udfmToUfm fvs, vectAvoidInfoOf eVI), AnnTick tick eVI)
+    ; return ((fvs, vectAvoidInfoOf eVI), AnnTick tick eVI)
     }
   where
     fvs = freeVarsOf ce
 
 vectAvoidInfo _pvs ce@(_, AnnType ty)
-  = return ((udfmToUfm fvs, VISimple), AnnType ty)
+  = return ((fvs, VISimple), AnnType ty)
   where
     fvs = freeVarsOf ce
 
 vectAvoidInfo _pvs ce@(_, AnnCoercion coe)
-  = return ((udfmToUfm fvs, VISimple), AnnCoercion coe)
+  = return ((fvs, VISimple), AnnCoercion coe)
   where
     fvs = freeVarsOf ce
 
@@ -1248,8 +1248,8 @@ allScalarVarType vs = and <$> mapM isScalarOrToplevel vs
 
 -- Are the types of all variables in the set in the 'Scalar' class or toplevel variables?
 --
-allScalarVarTypeSet :: VarSet -> VM Bool
-allScalarVarTypeSet = allScalarVarType . varSetElems
+allScalarVarTypeSet :: DVarSet -> VM Bool
+allScalarVarTypeSet = allScalarVarType . dVarSetElems
 
 -- Debugging support
 --
